@@ -1,6 +1,6 @@
 from perceiver.data.text.bookcorpus import BookCorpusDataModule
-from perceiver.data.text.common import TextPreprocessor
+from perceiver.data.text.common import Task, TextPreprocessor
+from perceiver.data.text.enwik8 import Enwik8DataModule
 from perceiver.data.text.imdb import ImdbDataModule
-from perceiver.data.text.wikibook import WikiBookDataModule
 from perceiver.data.text.wikipedia import WikipediaDataModule
 from perceiver.data.text.wikitext import WikiTextDataModule
