@@ -1,9 +1,7 @@
-import s3fs
-
 from perceiver.data.text import (
     BookCorpusDataModule,
+    Enwik8DataModule,
     ImdbDataModule,
-    WikiBookDataModule,
     WikipediaDataModule,
     WikiTextDataModule,
 )
