@@ -1,12 +1,12 @@
 import multiprocessing as mp
-from typing import Optional
 
 import jsonargparse
 
 from perceiver.data.text import (
     BookCorpusDataModule,
+    Enwik8DataModule,
     ImdbDataModule,
-    WikiBookDataModule,
+    Task,
     WikipediaDataModule,
     WikiTextDataModule,
 )
@@ -14,25 +14,30 @@ from perceiver.data.text import (
 
 DATAMODULE_CLASSES = {
     "bookcorpus": BookCorpusDataModule,
-    "wikipedia": WikipediaDataModule,
-    "wikibook": WikiBookDataModule,
-    "wikitext": WikiTextDataModule,
+    "enwik8": Enwik8DataModule,
     "imdb": ImdbDataModule,
+    "wikipedia": WikipediaDataModule,
+    "wikitext": WikiTextDataModule,
 }
 
 
 def main(args):
-    DATAMODULE_CLASSES[args.dataset](**args).prepare_data()
+    dm_class = DATAMODULE_CLASSES[args.dataset]
+
+    del args.dataset
+
+    dm_class(**args, pin_memory=False).prepare_data()
 
 
 if __name__ == "__main__":
     parser = jsonargparse.ArgumentParser(description="Preprocess dataset for training")
     parser.add_argument("dataset", default="wikitext", choices=list(DATAMODULE_CLASSES.keys()))
-    parser.add_argument("--tokenizer")
-    parser.add_argument("--max_seq_len", default=2048, type=int)
+    parser.add_argument("--tokenizer", type=str)
+    parser.add_argument("--max_seq_len", type=int)
+    parser.add_argument("--task", default=Task.mlm, type=Task)
+    parser.add_argument("--mask_prob", default=0.15, type=float)
+    parser.add_argument("--mask_words", default=True, type=bool)
+    parser.add_argument("--static_masking", default=False, type=bool)
     parser.add_argument("--add_special_tokens", default=False, type=bool)
-    parser.add_argument("--config_name", type=Optional[str])  # wikitext only
-    parser.add_argument("--filter_empty", default=True, type=bool)  # wikitext only
-    parser.add_argument("--filter_headers", default=False, type=bool)  # wikitext only
-    parser.add_argument("--num_workers", default=mp.cpu_count(), type=int)
+    parser.add_argument("--preproc_workers", default=mp.cpu_count(), type=int)
     main(parser.parse_args())
